%% Source code for the following paper:
% 
%  Chau-Wai Wong, Guan-Ming Su, Min Wu, "Impact Analysis of Baseband
%  Quantizer on Coding Efficiency for HDR Video," IEEE Signal Processing
%  Letters.


%% This simulation program calculates parameters gamma1 and gamma12
%  
%  Last modified by Chau-Wai Wong on 2/29/2016
%
%  Copyright (c) 2016  Chau-Wai Wong
%  Licensed under MIT License. For the full license, refer to LICENSE.txt


%% Parameters setup

blkSize = 8;  % transform size

q2q1ratio_range = 1:0.1:5;  % list of q2/q1 ratios to be used

% Size/length parameters for simulation. Ususally the larger, the more
% precise. For a quick check, set (M, count, repeatCnt) to (10, 10, 20);
% recommended values: (100, 100, 200).
M = 10;
count = 10;
repeatCnt = 20; % increasing this value improves the estimates significantly


%% Main routine for calculating gamma1 and gamma12

roundResidue = @(x) round(x) - x;

vecLen = blkSize^2;

plot_result_alpha = [];
plot_result_alpha_Y = [];
plot_result_aplha_W = [];

T_1d = dct(eye(blkSize));   % We use DCT (a non-degenerate transform) here,
%T_1d = myWht(blkSize);     % but other transforms can be used.

T = kron(T_1d, T_1d);

for q2q1ratio = q2q1ratio_range
    disp(q2q1ratio)

    alpha_collection = zeros(repeatCnt, 1);
    SSE_Y_collection = zeros(repeatCnt, 1);
    SSE_W_collection = zeros(repeatCnt, 1);
    
    % Can replace "parfor" by "for" if parallel computing capability is not
    % available.
    parfor repeat = 1 : repeatCnt
        
        randIdxSet_all = round((rand(vecLen, count)*2 - 1) * M);
        
        SSE_Y_arr = [];
        SSE_W_arr = [];
        W_cross_Y_arr = zeros(vecLen, vecLen, count);
        for i = 1 : count
            idxVec = randIdxSet_all(:,count);
            Y_vector = roundResidue ( T * idxVec / q2q1ratio) ;
            W_vector = roundResidue ( idxVec + q2q1ratio * (T') * Y_vector );
            W_cross_Y = W_vector * (Y_vector');
            SSE_Y = sum(Y_vector.^2);
            SSE_W = sum(W_vector.^2);
            
            W_cross_Y_arr(:,:,i) = W_cross_Y;
            SSE_Y_arr(i) = SSE_Y;
            SSE_W_arr(i) = SSE_W;
        end
        
        alpha_collection(repeat) = sum(diag( T * mean(W_cross_Y_arr, 3) ));
        SSE_Y_collection(repeat) = mean(SSE_Y_arr);
        SSE_W_collection(repeat) = mean(SSE_W_arr);
    end

    plot_result_alpha(end+1) = mean(alpha_collection*12/vecLen);

    plot_result_alpha_Y(end+1) = mean(SSE_Y_collection*12/vecLen);

    plot_result_aplha_W(end+1) = mean(SSE_W_collection*12/vecLen);
    
end

figure;
plot(q2q1ratio_range, plot_result_alpha, q2q1ratio_range, plot_result_alpha_Y, q2q1ratio_range, plot_result_aplha_W)
ylim([-0.6 1.2])
xlabel('q_2/q_1'); ylabel('estimates');
legend('\gamma_{12}', '\gamma_{2}', '\gamma_{1}')

%% save data

res.t = q2q1ratio_range;
res.gamma12 = plot_result_alpha;
res.gamma2 = plot_result_alpha_Y;
res.gamma1 = plot_result_aplha_W;
%save('gamma_N265_dct.mat', 'res');


%% display gamma vs. alpha

hFig = figure(9); set(hFig, 'position', [50 50 400 250]);
%load('gamma_N64_dct.mat');
load('gamma_N265_dct.mat');
plot(res.t, res.gamma1, 'b-', res.t, res.gamma12, 'r-.')
ylim([-0.5 1.2])
xlabel('q_2/q_1 or \alpha'); ylabel('estimates');
legend('\gamma_{1}', '\gamma_{12}', 'location', 'east', 'Interpreter', 'LaTex')
grid on; grid minor


%% display SNR loss vs. alpha

hFig = figure(10); set(hFig, 'position', [50 50 400 250]);
setRight = res.t > 2;
setLeft = ~setRight;
result_right = 10*log10(1 + 2 ./ (res.t(setRight).^2));
result_left = 10*log10(1 + (1+res.gamma1(setLeft)) ./ (res.t(setLeft).^2) + 2*res.gamma12(setLeft) ./ res.t(setLeft));
plot(res.t, [result_left result_right], 'k');
xlabel('q_2/q_1 or \alpha'); ylabel('SNR LOSS (dB)');
grid on; grid minor

