%% Source code for the following paper:
% 
%  Chau-Wai Wong, Guan-Ming Su, Min Wu, "Impact Analysis of Baseband
%  Quantizer on Coding Efficiency for HDR Video," IEEE Signal Processing
%  Letters.

%% Simulation for a transform codec with baseband quantization
%  [non-predictive coding branch of Fig. 1(b)]
%  
%  Pipeline structure:
% 
%    input --> [Q1]--[iQ1]--[T]--[Q2]--[iQ2]--[T]--[Round]--[Q1]--[iQ1]--> output
%  
%  Last modified by Chau-Wai Wong in Dec. 2016
%
%  Copyright (c) 2016  Chau-Wai Wong
%  Licensed under MIT License. For the full license, refer to LICENSE.txt


%% Parameters setup

% Below are two possible configurations for the simulation and 
% corresponding results:
%
% (   4^2   ,   0.4   ): 2.9999 1.7416 0.5083 0.1285
% (  16^2   ,   0.9   ): 2.9544 1.7530 0.5334 0.1567
%      ^         ^                     ^
%   vecLen    rho_ar1    4 SNR difference values at the same bitrate

sourceInfo = struct ( 'count'     , 10000, ... % number of Monte Carlo sample points
                      'vecLen'    , 16^2 , ... % transform size
                      'rho_ar1'   , 0.9  , ... % AR(1) parameter
                      'trueRandom', 0    );    % pseudo/true random flag


%% Simulation

% Standard deviation of AR(1) process with unit-variance innovation
std_theory = sqrt(1/(1-sourceInfo.rho_ar1^2))

quantizer2ptsCnt = 4;    % number of different quantization levels for drawing rate-distortion (RD) curves
quantizer2Steps = 2.^[0:quantizer2ptsCnt-1] * std_theory/10;

q_step1_coarse  = min(quantizer2Steps); 
q_step1_precise = q_step1_coarse/10;

% Initialize variables
bitrateDistribution_precise = zeros(sourceInfo.vecLen, quantizer2ptsCnt);
bitrateDistribution_coarse  = zeros(sourceInfo.vecLen, quantizer2ptsCnt);
dist_precise    =  zeros(1, quantizer2ptsCnt);
dist_coarse     =  zeros(1, quantizer2ptsCnt);
bitrate_precise =  zeros(1, quantizer2ptsCnt);
bitrate_coarse  =  zeros(1, quantizer2ptsCnt);

% Carry out simulations for different quantization strengths
for i = 1 : quantizer2ptsCnt
    q_step2 = quantizer2Steps(i);
    [dist_precise(i), bitrate_precise(i), bitrateDistribution_precise(:,i)] = func_simu3(q_step1_precise, q_step2, sourceInfo);
    [dist_coarse(i),  bitrate_coarse(i) , bitrateDistribution_coarse(:,i) ] = func_simu3(q_step1_coarse , q_step2, sourceInfo);
end

% Calculate Signal-to-Noise Ratios (SNRs)
snr_precise = 10*log10(std_theory^2./dist_precise);
snr_coarse  = 10*log10(std_theory^2./dist_coarse);


%% Display results
hFig = figure(9); set(hFig, 'position', [50 50 400 250]);
%plot(bitrate_precise, snr_precise, '-o', bitrate_coarse, snr_coarse, '--o');

% fit smooth curves to the RD points 
[xData, yData] = prepareCurveData( bitrate_precise, snr_precise );
[fitresult_precise, gof] = fit( xData, yData, 'splineinterp', 'Normalize', 'on' );
plot( fitresult_precise, 'b', xData, yData, ['b*']); hold on

[xData, yData] = prepareCurveData( bitrate_coarse, snr_coarse );
[fitresult_coarse, gof] = fit( xData, yData, 'splineinterp', 'Normalize', 'on' );
plot( fitresult_coarse, 'r-.', xData, yData, ['r+']); hold off

xlabel('bitrate (bits/pixel)'); ylabel('SNR (dB)'); grid on; grid minor;

legend_q1 = 'fitted curve, $q_1 \to 0$';
legend_q2 = ['fitted curve, $q_1 = ' sprintf('%.2f', q_step1_coarse) '$'];
legend({'precise $Q_1$', legend_q1, 'coarse $Q_1$', legend_q2}, 'location', 'northwest', 'Interpreter', 'LaTex');

% show legend
yRange = 10:32;
hold on;
for i = 1:4
    thisBitrate = bitrate_precise(i);
    plot(thisBitrate*ones(1,length(yRange)), yRange, 'k--');
    q2_text = sprintf('$q_2 = %.2f$', quantizer2Steps(i));
    text(thisBitrate-0.3, 11, q2_text, 'Interpreter', 'LaTex');
end
hold off;

% evaluate SNR (y-axis) values at particular bitrate (x-axis) values
snr_precise_hat = feval(fitresult_precise, bitrate_coarse);

% calculate SNR difference at the same bitrate (aka Bjontegaard SNR change)
deltaSnr = snr_precise_hat(:) - snr_coarse(:)

